(*
This file is part of the Kind verifier

* Copyright (c) 2007-2011 by the Board of Trustees of the University of Iowa, 
* here after designated as the Copyright Holder.
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of the University of Iowa, nor the
*       names of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER ''AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*)


(** This module provides logging support for Kind2:
    - logging to a channel (stderr, stdout, ...)
    - logging to a file
*)


(** 

    @author Corey Oliver
    @author Temesghen Kahsai

*)


(** Type of log levels. Below they are defined in order of decreasing
    severity. *)
type level =
    SEVERE
      (** Indicates a serious failure. *)
  | WARNING
      (** Indicates a potential problem. *)
  | INFO
      (** Used of information messages. *)
  | DEBUG
      (** Debugging message. *)

type formatter = out_channel -> level -> float -> string -> unit
(** [formatter channel level timestamp message] formats a log
    message.*)

type logger

val level_to_string : level -> string
(** Converts a level to its string equivalent. *)

val format_timestamp : out_channel -> float -> unit
(** Formats a timestamp. *)

val std_formatter : out_channel -> level -> float -> string -> unit
(** An example of a formatter. *)
  
val mk_logger : level -> formatter -> out_channel -> bool -> open_flag list -> int -> string -> logger
(** [mk_logger] returns a new logger. *)
  
val log : logger -> ?level : level -> ?timestamp : float -> string -> unit
(** [log logger ?level ?timestamp message] logs a message. *)
  
val set_level : logger -> level -> unit
(** Set the log level specifying which message levels will be
    logged. *)

val set_formatter : logger -> formatter -> unit
(** Set the formatter used by the logger. *)

val set_channel : logger -> out_channel -> unit
(** Set the channel used by the logger. *)

val set_write_to_file : logger -> bool -> unit
(** Set whether the logger prints to a file *)

val set_file_mode : logger -> open_flag list -> unit
(** Set the file modes which the logger use to write to a file. *)

val set_file_perm : logger -> int -> unit
(** Set the file permission used by the logger to write to a file. *)
  
val set_file_name : logger -> string -> unit
(** Set the file name the logger writes to (if [file] is true). *)
