(*
This file is part of the Kind verifier

* Copyright (c) 2007-2012 by the Board of Trustees of the University of Iowa, 
* here after designated as the Copyright Holder.
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of the University of Iowa, nor the
*       names of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER ''AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*)


(** Incremental and Parallel Kind for verification of multiple properties.
    A front end that allow you to interact with pkind without going trough mpiexec. 
*)


(** 
@author Temesghen Kahsai

*)

open Types
open Exceptions
open Channels
open Globals


let solver = Globals.my_solver

(********************************************************************)
let print_copyright () =
  print_string (
   "\nParallel K-Induction-based model checker (version "^Globals.version
  ^", build "^Globals.build^")\n"
  ^"* Copyright (c) 2007-2012 by the Board of Trustees of the University of Iowa.\n"
  ^"* All rights reserved.\n"
  ^"* \n"
  ^"* Redistribution and use in source and binary forms, with or without\n"
  ^"* modification, are permitted provided that the following conditions are met:\n"
  ^"*     * Redistributions of source code must retain the above copyright\n"
  ^"*       notice, this list of conditions and the following disclaimer.\n"
  ^"*     * Redistributions in binary form must reproduce the above copyright\n"
  ^"*       notice, this list of conditions and the following disclaimer in the\n"
  ^"*       documentation and/or other materials provided with the distribution.\n"
  ^"*     * Neither the name of the University of Iowa, nor the\n"
  ^"*       names of its contributors may be used to endorse or promote products\n"
  ^"*       derived from this software without specific prior written permission.\n"
  ^"*\n"
(*  
  ^"* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER ''AS IS'' AND ANY\n"
  ^"* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED\n"
  ^"* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE\n"
  ^"* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY\n"
  ^"* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES\n"
  ^"* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;\n"
  ^"* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND\n"
  ^"* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT\n"
  ^"* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS\n"
  ^"* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.\n"
*)
  )

let pkind_cmd = "pkind-main"  (* pkind binary*)

(********************************************************************)

(* Various options *)

let inv_gen = ref false
let kind_ai = ref false
let arg_node : string option ref = ref None
let arg_n: int option ref = ref None
let arg_no_multi = ref false
let arg_timeout : float option ref = ref None
let arg_k_incremental : int option ref = ref None
let arg_no_incremental = ref false
let arg_range : int option ref = ref None
let arg_no_mode_inv = ref false
let arg_mpi_abort = ref false
let arg_cvc3 = ref false
let arg_cvc4 = ref false
let arg_yices_exec= ref false
let arg_solver_path: string option ref = ref None
let arg_compression = ref false
let arg_loop = ref false
let arg_flags: string option ref = ref None
let arg_xml = ref false
let arg_quiet = ref false
let arg_verbose = ref false
let arg_induct_cex = ref false
let arg_scratch = ref false
let arg_only_inputs = ref false
let arg_dot = ref false
let arg_dotall = ref false
let arg_version = ref false
let arg_no_inlining = ref false
let arg_no_imp = ref false
let arg_naive = ref false

let get_args () =
  let list =
    (if !inv_gen then ["-with-inv-gen"] else []) @
      (if !kind_ai then ["-with-kind-ai"] else []) @ 
      (match !arg_node with None -> [] | Some s -> ["-node " ^ s]) @
      (match !arg_n with None -> [] | Some i -> ["-n " ^ string_of_int i]) @
      (if !arg_no_multi then ["-no_multi"] else []) @
      (match !arg_timeout with None -> [] | Some x -> ["-timeout " ^ string_of_float x]) @
      (match !arg_k_incremental with None -> [] | Some x -> ["-k_incremental " ^ string_of_int x]) @
      (if !arg_no_incremental then ["-no_incremental"] else []) @
      (match !arg_range with None -> [] | Some x -> ["-range " ^ string_of_int x]) @
      (if !arg_no_mode_inv then ["-no_mode_inv"] else []) @
      (if !arg_mpi_abort then ["-mpi_abort"] else []) @
      (if !arg_cvc3 then ["-cvc3"] else []) @
      (if !arg_cvc4 then ["-cvc4"] else []) @
      (if !arg_yices_exec then ["-yices_exec"] else []) @
      (match !arg_solver_path with None -> [] | Some x -> ["-solver_path " ^ x]) @
      (if !arg_compression then ["-compression"] else []) @
      (if !arg_loop then ["-loop"] else []) @
      (match !arg_flags with None -> [] | Some x -> ["-flags " ^ x]) @
      (if !arg_xml then ["-xml"] else []) @
      (if !arg_quiet then ["-quiet"] else []) @
      (if !arg_verbose then ["-verbose"] else []) @
      (if !arg_induct_cex then ["-induct_cex"] else []) @
      (if !arg_scratch then ["-scratch"] else []) @
      (if !arg_only_inputs then ["-only_inputs"] else []) @
      (if !arg_dot then ["-dot"] else []) @
      (if !arg_dotall then ["-dotall"] else []) @ 
      (if !arg_no_inlining then ["-no_inlining"] else []) 
  in
  Array.of_list list 
      
let start fn =
  print_to_user_final ("\n\n==========================================\n");
  print_to_user_final     ("  PARALLEL KIND                           \n");
  if !kind_ai then
    print_to_user_final   ("               * WITH KIND-AI INVARIANTS  \n");
  if !inv_gen then
    print_to_user_final   ("               * WITH INVARIANT GENERATION\n");
  print_to_user_final     ("==========================================\n\n");
  if !arg_xml then 
    print_to_user_final ("  KIND results will be saved in: " ^ fn ^ ".xml\n");

  let nb_proc = 
    match !inv_gen, !kind_ai with
      | true, true -> 4
      | false, true 
      | true, false -> 3
      | false, false -> 2 
  in
  let arg = Array.concat [[| "mpiexec" ; "-n"; string_of_int nb_proc;  pkind_cmd |]; get_args (); [|fn|]] in 
  let cmd = Array.fold_left (fun res s -> res ^ " " ^ s) "" arg in
  let (in_chan, out_chan) as proc = Unix.open_process cmd in
  try
    while true do print_endline (input_line in_chan); done
  with End_of_file -> (
    let proc_status = Unix.close_process proc in
    exit 5646)
    

let set refvar s = refvar := Some s 
(********************************************************************)
let _ =
  let usage_msg = "PKIND -- Incremental and Parallel k-induction model checker for multiple properties Lustre programs. \nUsage: pkind [options] input_file" in
  let speclist = [
		   ("-with-inv-gen",
                    Arg.Set inv_gen,
                    "\tCONTROL: Activate invariant generation  based on a predefined template. It will generate invariants of type BOOL and INT."
                   );

		   ("-with-kind-ai",
                    Arg.Set kind_ai,
                    "\tCONTROL: Activate invariant generation based on abstract interpretation. KIND-AI  must be installed and configures with './configure'."
                   );

                   ("-node", Arg.String (set arg_node),
		     "\tCONTROL: User specified main node (default: the last node that contains a PROPERTY)");

                   ("-n", Arg.Int (set arg_n),
                    "\tCONTROL: Stop after n iterations (default 200)"
                   );

		   ("-no_multi", Arg.Set arg_no_multi,
                    "\tCONTROL: Disables the verification of multiple properties in an incremental fascion. (default: true)"
		   );

		   ("-timeout", Arg.Float (set arg_timeout),
                    "\tCONTROL: For setting a timeout (default: 100.00)"
                   );
         		   
		   ("-k_incremental", Arg.Int (set arg_k_incremental),
                    "\tINVARIANT GENERATION: Generate incremental invariants until n iterations (default: k=200)"
                   );
		   
		   ("-no_incremental", Arg.Set arg_no_incremental,
                    "\tINVARIANT GENERATION: Generate non incremental invariants. (default: generate incremental invariants)"
                   );

		   ("-range", Arg.Int (set arg_range),
		    "\tINVARIANT GENERATION: Maximum range for which to generate candidates for mode variables (default: max range = 10)");
		   
		   ("-no_mode_inv",
                    Arg.Set arg_no_mode_inv,
                    "\tINVARIANT GENERATION: Don not generate invariants for mode variables. (default: Generate invariants for mode variables)"
                   );

                   ("-mpi_abort",
                    Arg.Set arg_mpi_abort,
                    "\tPARALLELISM: It will abort the mpi environment. Useful when the solver returns with an answer but the environment is not terminated yet. (default false)"
                   );
                   
		   ("-cvc3",
                    Arg.Set arg_cvc3,
                    "\tSOLVER: Use CVC3 as solver (default: yicesw)"
                   );
		   
		   ("-cvc4",
                    Arg.Set arg_cvc4,
                    "\tSOLVER: Use CVC4 as solver [Work in progress] (default: yicesw)"
                   );
                   
		   ("-yices_exec",
                    Arg.Set arg_yices_exec,
                    "\tSOLVER: Use Yices executable as solver (default: yices)"
                   );
  
		   ("-solver_path",
                    Arg.String (set arg_solver_path),
                    "\tSOLVER: Specify the path to the solver. Used for KIND GUI. (default: specified by ./configure)"
                   );

		   ("-no_inlining",
		    Arg.Set arg_no_inlining,
                    "\tOPT: Do not attempt variable inlining (default: false)"
                   );
         
		   ("-compression",
                    Arg.Set arg_compression,
                    "\tCOMPR: Include path compression but not term check (default: false)"
                   );
		   
		   ("-loop",
                    Arg.Set arg_loop,
                    "\tCOMPR: Include loopfree constraint formulas + term check (default: false)"
                   );
		   
		   ("-flags",
                    Arg.String (set arg_flags),
                    "\tSOLVER: Pass string to solver cammand line (default: n/a)"
                   );

		   ("-xml",		    Arg.Set arg_xml,
                    "\tOUTPUT: Generate KIND result in XML format (default: false)"
                   );
		   
                   ("-quiet",                    Arg.Set arg_quiet,
                    "\tOUTPUT: Only return final answer (default: true)"
                   );
                   
		   ("-verbose",
                    Arg.Set arg_verbose,
                    "\tOUTPUT: Return more details (default: false)"
                   );

		   ("-induct_cex",
                    Arg.Set arg_induct_cex,
                    "\tOUTPUT: Prints the current inductive counterexample (default: false)"
                   );

		    ("-scratch",                    Arg.Set arg_scratch,
                    "\tOUTPUT: Produce files for debugging purposes. This option will produce 3 files (*_base, *_induct, *_inv). (default: Don't log to disk)"
                   );

                   ("-only_inputs",                    Arg.Set arg_only_inputs,
                    "\tOUTPUT: Only print inputs in counterexamples (default: false)"
                   );

                   ("-dot",                   Arg.Set arg_dot,
                    "\tOUTPUT: Prints dot graph of dependencies (default: false)"
                   );

                   ("-dotall",
                    Arg.Set arg_dotall,
                    "\tOUTPUT: Outputs all abstraction dependency graphs to graphXXX.dot (default: false)"
                   );  
    
                   ("-version",
		    Arg.Unit (fun () -> print_copyright()),
                      "Print version & copyright information"
                   )
                 ]
  in
  Arg.parse speclist start usage_msg 


